#include <Servo.h>
#include <spi.h>

#define DEBUG true
#define STATE_KEYBOARD 0
#define STATE_RF 1
#define STATE_WEB_SERIAL 2
#define STATE_WEB_RF 3

#define DEFAULT_STATE STATE_WEB_SERIAL

#define PIN_SERVO_TAIL 9
#define PIN_MOTOR_TAIL 5
#define PIN_MOTOR_CENTER 6

#define SERVO_MIN -100
#define INPUT_MIN 0
#define INPUT_MAX 100

#define SERVO_TAIL_CENTER_ANGLE 94
#define SERVO_TAIL_MIN_ANGLE 0
#define SERVO_TAIL_MAX_ANGLE 188

#define MOTOR_TAIL_MIN_POWER 20
#define MOTOR_TAIL_MAX_POWER 180
#define MOTOR_CENTER_MIN_POWER 50
#define MOTOR_CENTER_MAX_POWER 130

#define BAUD_RATE 9600
#define CHAR_NEWLINE 10
#define BUFFER_SIZE 128

#define SWITCH_STATES 'c'
#define TYPE_MOTOR 'm'
#define TYPE_SERVO 't'

#define CENTER_MOTOR_FASTER 'w'
#define CENTER_MOTOR_SLOWER 's'
#define CENTER_MOTOR_STOP '1'
#define TAIL_MOTOR_FASTER 'i'
#define TAIL_MOTOR_SLOWER ','
#define TAIL_MOTOR_STOP '2'
#define TAIL_SERVO_LEFT 'j'
#define TAIL_SERVO_RIGHT 'l'
#define TAIL_SERVO_CENTER 'k'
#define STOP_ALL_MOTORS ' '

#define MOTOR_STEP 10
#define SERVO_STEP 5

#define TYPE_INDEX 0
#define NUMBER_INDEX 1
#define VALUE_INDEX 3

// type[# value]

// examples:
// m1 90 (motor 1, set power to 90)
// s1 90 (servo 1, set angle to 90)
// c (switch states)

Servo tailServo;
Servo tailMotor;
Servo centerMotor;

int state;
char serialInput[BUFFER_SIZE];
int serialIndex; 
char tmpChar;

int centerMotorPower;
int tailMotorPower;
int tailServoAngle;

char buffer[BUFFER_SIZE];
int status;
boolean readingSerial;

/**
 * Initialize
 **/
void setup() {
  Serial.begin(BAUD_RATE);
  SPI_initSlave();
  
  tailServo.attach(PIN_SERVO_TAIL);
  tailMotor.attach(PIN_MOTOR_TAIL);
  centerMotor.attach(PIN_MOTOR_CENTER);

  // set default values
  state = DEFAULT_STATE;
  serialIndex = 0;
  centerMotorPower = MOTOR_CENTER_MIN_POWER;
  tailMotorPower = MOTOR_TAIL_MIN_POWER;
  tailServoAngle = SERVO_TAIL_CENTER_ANGLE;
  readingSerial = false;
  
  // make sure the engines are off
  tailMotor.write(tailMotorPower);
  centerMotor.write(centerMotorPower);

  // set default starting angle for the tail
  tailServo.write(tailServoAngle);
 
  
  if (DEBUG) {
    Serial.println("initialization complete");
    Serial.println("");
    Serial.println("Protocol:");
    Serial.println("type[# value]");
    Serial.println("");
    Serial.println("examples:");
    Serial.println("m1 90 (motor 1, set power to 90)");
    Serial.println("s1 90 (servo 1, set angle to 90)");
    Serial.println("c (switch states)");
  }
}

/**
 * Main loop
 **/
void loop() {
  switch (state) {
    case STATE_KEYBOARD:
    case STATE_WEB_SERIAL:
      handleKeyboard();
      break;
      
    case STATE_WEB_RF:
      handleWebRF();
      break;
      
    case STATE_RF:
      handleRF();
      break;
  }
}

/**
 * Handle incoming commands
 *
 * @param *input
 * @param size The length of input
 */
void handleInputCommand(char *input, int size) {
  if (size == 0) {
    return;
  }
  
  char type = input[TYPE_INDEX];
  if (DEBUG) {
    Serial.print("Type: ");
    Serial.println(type);
  }
  
  if (type == SWITCH_STATES) {
    // change states
    state = STATE_RF;
    if (DEBUG) {
      Serial.println("-- switching to RF state");
    } 
  } else if (size > VALUE_INDEX) {
    boolean protocolPass = false;
    Servo outputDevice;
    int value = atoi(input + VALUE_INDEX);
    
    // select output device
    if (type == TYPE_MOTOR) {
      protocolPass = true;
      if (input[NUMBER_INDEX] == '1') {
        outputDevice = tailMotor;
        value = constrain(map(value, INPUT_MIN, INPUT_MAX, MOTOR_TAIL_MIN_POWER, MOTOR_TAIL_MAX_POWER), MOTOR_TAIL_MIN_POWER, MOTOR_TAIL_MAX_POWER);
        tailMotorPower = value;
        if (DEBUG) {
          Serial.println("-- tail motor selected");
        }
      } else {
        outputDevice = centerMotor;
        value = constrain(map(value, INPUT_MIN, INPUT_MAX, MOTOR_CENTER_MIN_POWER, MOTOR_CENTER_MAX_POWER), MOTOR_CENTER_MIN_POWER, MOTOR_CENTER_MAX_POWER);
        centerMotorPower = value;
        if (DEBUG) {
          Serial.println("-- center motor selected");
        }
      }
    } else if (type == TYPE_SERVO) {
      protocolPass = true;
      outputDevice = tailServo;
      value = constrain(map(value, SERVO_MIN, INPUT_MAX, SERVO_TAIL_MIN_ANGLE, SERVO_TAIL_MAX_ANGLE), SERVO_TAIL_MIN_ANGLE, SERVO_TAIL_MAX_ANGLE);
      tailServoAngle = value;
      if (DEBUG) {
        Serial.println("-- tail servo selected");
      }
    }
    
    if (protocolPass) {
      // set output device value
      if (DEBUG) {
        Serial.print("value: ");
        Serial.println(value);
      }
      
      outputDevice.write(value);
    } else if (DEBUG) {
      Serial.println("bad input format");
    }
  }  
}

/**
 * Parse serial input and pass it forward to the handling module
 **/
void handleKeyboard() {  
  if (Serial.available() > 0) {      
      tmpChar = Serial.read();
      handleChar(tmpChar);
  }
}

/**
 * Parse radio input and pass it forward to the handling module
 **/
void handleRF() {  
 if (SPI_isReceiveReady()) {
      status = SPI_receive(buffer, BUFFER_SIZE);
      
      // be forgiving, we're broadcasting the same character multiple times
      // so chances are we got one right
      if (DEBUG) {
          Serial.print("received: ");
          Serial.println(buffer);  
       }
        
      handleChar(buffer[0]);
      
//      if (status == SUCCESS) {
//          if (DEBUG) {
//            Serial.print("received: ");
//            Serial.println(buffer);  
//          }
//          
//          handleChar(buffer[0]);
//      } else {
//          Serial.println("transmission failure!");
//      }
  }
}

/**
 * Parse radio input and pass it forward to the handling module
 **/
void handleWebRF() {  
 if (SPI_isReceiveReady()) {
      status = SPI_receive(buffer, BUFFER_SIZE);
      if (status == SUCCESS) {
        if (DEBUG) {
          Serial.print("received: ");
          Serial.println(buffer);
        }
        
        handleInputCommand(buffer, BUFFER_SIZE);
      } else if (DEBUG) {
        Serial.println("transmission failure!");
      }

  }
}

void handleChar(char c) {
  switch (c) {
        case CHAR_NEWLINE:
          readingSerial = false;
          handleInputCommand(serialInput, serialIndex);
        
          // clear saved input
          serialIndex = 0;
          serialInput[serialIndex] = '\0';
          break;
        
        case CENTER_MOTOR_FASTER:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (centerMotorPower < MOTOR_CENTER_MAX_POWER) {        
            if (DEBUG) {
              Serial.println("- center motor faster");
            }
            
            centerMotorPower += MOTOR_STEP;
            centerMotor.write(centerMotorPower);
          }
          
          break;
          
        case CENTER_MOTOR_SLOWER:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
            
          if (centerMotorPower > MOTOR_CENTER_MIN_POWER) {        
            if (DEBUG) {
              Serial.println("- center motor slower");
            }
            
            centerMotorPower -= MOTOR_STEP;
            centerMotor.write(centerMotorPower);
          }
          
          break;
          
        case CENTER_MOTOR_STOP:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (DEBUG) {
            Serial.println("- center motor stop");
          }
          
          centerMotorPower = MOTOR_CENTER_MIN_POWER;
          centerMotor.write(centerMotorPower);
          break;
        
        case TAIL_MOTOR_FASTER:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (tailMotorPower < MOTOR_TAIL_MAX_POWER) {
            if (DEBUG) {
              Serial.println("- tail motor faster");
            }       
            
            tailMotorPower += MOTOR_STEP;
            tailMotor.write(tailMotorPower);
          }
          
          break;
          
        case TAIL_MOTOR_SLOWER:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (tailMotorPower > MOTOR_TAIL_MIN_POWER) {    
            if (DEBUG) {
              Serial.println("- tail motor slower");
            }    
            
            tailMotorPower -= MOTOR_STEP;
            tailMotor.write(tailMotorPower);
          }
          
          break;

        case TAIL_MOTOR_STOP:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
            
          if (DEBUG) {
            Serial.println("- tail motor stop");
          }
          
          tailMotorPower = MOTOR_TAIL_MIN_POWER;
          tailMotor.write(tailMotorPower);
          break;
          
        case TAIL_SERVO_LEFT:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (tailServoAngle > SERVO_TAIL_MIN_ANGLE) {        
            if (DEBUG) {
              Serial.println("- tail servo left");
            }
            
            tailServoAngle -= SERVO_STEP;
            tailServo.write(tailServoAngle);
          }
          
          break;
        case TAIL_SERVO_RIGHT:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
            
          if (tailServoAngle < SERVO_TAIL_MAX_ANGLE) {        
            if (DEBUG) {
              Serial.println("- tail servo right");
            }
            
            tailServoAngle += SERVO_STEP;
            tailServo.write(tailServoAngle);
          }
          
          break;
          
        case TAIL_SERVO_CENTER:
            if (readingSerial) {
              handleDefaultKey(c);
              break;
            }
            
            if (DEBUG) {
              Serial.println("- tail servo center");
            }
            
            tailServoAngle = SERVO_TAIL_CENTER_ANGLE;
            tailServo.write(tailServoAngle);
        
          break;
          
        case STOP_ALL_MOTORS:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (DEBUG) {
            Serial.println("** STOPPING ALL MOTORS **");
          }
          
          centerMotorPower = MOTOR_CENTER_MIN_POWER;
          centerMotor.write(centerMotorPower);
          tailMotorPower = MOTOR_TAIL_MIN_POWER;
          tailMotor.write(tailMotorPower);          
          break;
          
        default:
            if (state != STATE_RF) {
              readingSerial = true;
              handleDefaultKey(c);
            }
      }  
}


void handleDefaultKey(char c) {  
  serialInput[serialIndex++] = c;
  serialInput[serialIndex] = '\0'; // null terminate the string
  
  if (serialIndex > BUFFER_SIZE - 2) {
    if (DEBUG) {
      if (Serial.available() > 0) {
        Serial.println("exceeded buffer size, ignoring some input!");
      }  
    }
    
    // clear buffer and saved input
    while (Serial.read() != -1);
    serialIndex = 0;
    serialInput[serialIndex] = '\0';  
  }
}
